from ._chat import Chat, chat_ui
from ._chat_normalize import message_content, message_content_chunk
from ._markdown_stream import MarkdownStream, output_markdown_stream

__all__ = [
    "Chat",
    "chat_ui",
    "MarkdownStream",
    "output_markdown_stream",
    "message_content",
    "message_content_chunk",
]



from __future__ import annotations

import sys
from functools import singledispatch

from htmltools import HTML, Tagifiable, TagList

from ._chat_normalize_chatlas import tool_request_contents, tool_result_contents
from ._chat_types import ChatMessage

__all__ = ["message_content", "message_content_chunk"]


@singledispatch
def message_content(message):
    """
    Extract content from various message types into a ChatMessage.

    This function uses `singledispatch` to allow for easy extension to support
    new message types. To add support for a new type, register a new function
    using the `@message_content.register` decorator.

    Parameters
    ----------
    message
        The message object to extract content from (e.g., ChatCompletion,
        BaseMessage, etc.).

    Note
    ----
    This function is implicitly called by `Chat.append_message()` to support
    handling of various message types. It is not intended to be called directly
    by users, but may be useful for debugging or advanced use cases.

    Returns
    -------
    ChatMessage
        A ChatMessage object containing the extracted content and role.

    Raises
    ------
    ValueError
        If the message type is unsupported.
    """
    if isinstance(message, (str, HTML)) or message is None:
        return ChatMessage(content=message)
    if isinstance(message, ChatMessage):
        return message
    if isinstance(message, dict):
        if "content" not in message:
            raise ValueError("Message dictionary must have a 'content' key")
        return ChatMessage(
            content=message["content"],
            role=message.get("role", "assistant"),
        )
    raise ValueError(
        f"Don't know how to extract content for message type {type(message)}: {message}. "
        "Consider registering a function to handle this type via `@message_content.register`"
    )


@singledispatch
def message_content_chunk(chunk):
    """
    Extract content from various message chunk types into a ChatMessage.

    This function uses `singledispatch` to allow for easy extension to support
    new chunk types. To add support for a new type, register a new function
    using the `@message_content_chunk.register` decorator.

    Parameters
    ----------
    chunk
        The message chunk object to extract content from (e.g., ChatCompletionChunk,
        BaseMessageChunk, etc.).

    Note
    ----
    This function is implicitly called by `Chat.append_message_stream()` (on every
    chunk of a message stream). It is not intended to be called directly by
    users, but may be useful for debugging or advanced use cases.

    Returns
    -------
    ChatMessage
        A ChatMessage object containing the extracted content and role.

    Raises
    ------
    ValueError
        If the chunk type is unsupported.
    """
    if isinstance(chunk, (str, HTML)) or chunk is None:
        return ChatMessage(content=chunk)
    if isinstance(chunk, ChatMessage):
        return chunk
    if isinstance(chunk, dict):
        if "content" not in chunk:
            raise ValueError("Chunk dictionary must have a 'content' key")
        return ChatMessage(
            content=chunk["content"],
            role=chunk.get("role", "assistant"),
        )
    raise ValueError(
        f"Don't know how to extract content for message chunk type {type(chunk)}: {chunk}. "
        "Consider registering a function to handle this type via `@message_content_chunk.register`"
    )


# ------------------------------------------------------------------
# Shiny tagifiable content extractor
# ------------------------------------------------------------------


@message_content.register
def _(message: Tagifiable):
    return ChatMessage(content=message)


@message_content_chunk.register
def _(chunk: Tagifiable):
    return ChatMessage(content=chunk)


# -----------------------------------------------------------------
# chatlas tool call display
# -----------------------------------------------------------------
try:
    from chatlas import ContentToolRequest, ContentToolResult, Turn
    from chatlas.types import Content, ContentText

    @message_content.register
    def _(message: Content):
        return ChatMessage(content=str(message))

    @message_content_chunk.register
    def _(chunk: Content):
        return message_content(chunk)

    @message_content.register
    def _(message: ContentText):
        return ChatMessage(content=message.text)

    @message_content_chunk.register
    def _(chunk: ContentText):
        return message_content(chunk)

    @message_content.register
    def _(chunk: ContentToolRequest):
        return ChatMessage(content=tool_request_contents(chunk))

    @message_content_chunk.register
    def _(chunk: ContentToolRequest):
        return message_content(chunk)

    @message_content.register
    def _(chunk: ContentToolResult):
        return ChatMessage(content=tool_result_contents(chunk))

    @message_content_chunk.register
    def _(chunk: ContentToolResult):
        return message_content(chunk)

    @message_content.register
    def _(message: Turn):
        contents = TagList()
        for x in message.contents:
            contents.append(message_content(x).content)
        return ChatMessage(content=contents)

    @message_content_chunk.register
    def _(chunk: Turn):
        return message_content(chunk)

    # N.B., unlike R, Python Chat stores UI state and so can replay
    # it with additional workarounds. That's why R currently has a
    # shinychat_contents() method for Chat, but Python doesn't.
except ImportError:
    pass

# ------------------------------------------------------------------
# LangChain content extractor
# ------------------------------------------------------------------

try:
    from langchain_core.messages import BaseMessage, BaseMessageChunk

    @message_content.register
    def _(message: BaseMessage):
        if isinstance(message.content, list):
            raise ValueError(
                "The `message.content` provided seems to represent numerous messages. "
                "Consider iterating over `message.content` and calling .append_message() on each iteration."
            )
        return ChatMessage(
            content=message.content,
            role="assistant",
        )

    @message_content_chunk.register
    def _(chunk: BaseMessageChunk):
        if isinstance(chunk.content, list):
            raise ValueError(
                "The `chunk.content` provided seems to represent numerous message chunks. "
                "Consider iterating over `chunk.content` and calling .append_message() on each iteration."
            )
        return ChatMessage(
            content=chunk.content,
            role="assistant",
        )
except ImportError:
    pass


# ------------------------------------------------------------------
# OpenAI content extractor
# ------------------------------------------------------------------

try:
    from openai.types.chat import ChatCompletion, ChatCompletionChunk

    @message_content.register
    def _(message: ChatCompletion):
        return ChatMessage(
            content=message.choices[0].message.content,
            role="assistant",
        )

    @message_content_chunk.register
    def _(chunk: ChatCompletionChunk):
        return ChatMessage(
            content=chunk.choices[0].delta.content,
            role="assistant",
        )
except ImportError:
    pass


# ------------------------------------------------------------------
# Anthropic content extractor
# ------------------------------------------------------------------

try:
    from anthropic.types import (  # pyright: ignore[reportMissingImports]
        Message as AnthropicMessage,
    )

    @message_content.register
    def _(message: AnthropicMessage):
        content = message.content[0]
        if content.type != "text":
            raise ValueError(
                f"Anthropic message type {content.type} not supported. "
                "Only 'text' type is currently supported"
            )
        return ChatMessage(content=content.text)

    # Old versions of singledispatch doesn't seem to support union types
    if sys.version_info >= (3, 11):
        from anthropic.types import (  # pyright: ignore[reportMissingImports]
            RawMessageStreamEvent,
        )

        @message_content_chunk.register
        def _(chunk: RawMessageStreamEvent):
            content = ""
            if chunk.type == "content_block_delta":
                if chunk.delta.type != "text_delta":
                    raise ValueError(
                        f"Anthropic message delta type {chunk.delta.type} not supported. "
                        "Only 'text_delta' type is supported"
                    )
                content = chunk.delta.text

            return ChatMessage(content=content)
except ImportError:
    pass


# ------------------------------------------------------------------
# Google content extractor
# ------------------------------------------------------------------

try:
    from google.generativeai.types.generation_types import (
        GenerateContentResponse,
    )

    @message_content.register
    def _(message: GenerateContentResponse):
        return ChatMessage(content=message.text)

    @message_content_chunk.register
    def _(chunk: GenerateContentResponse):
        return ChatMessage(content=chunk.text)

except ImportError:
    pass


# ------------------------------------------------------------------
# Ollama content extractor
# ------------------------------------------------------------------

try:
    from ollama import ChatResponse

    @message_content.register
    def _(message: ChatResponse):
        msg = message.message
        return ChatMessage(msg.content)

    @message_content_chunk.register
    def _(chunk: ChatResponse):
        msg = chunk.message
        return ChatMessage(msg.content)

except ImportError:
    pass



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, TypedDict

from htmltools import HTML, Tag, TagChild, Tagifiable, TagList
from shiny.session import get_current_session

from ._typing_extensions import NotRequired

Role = Literal["assistant", "user", "system"]


# TODO: content should probably be [{"type": "text", "content": "..."}, {"type": "image", ...}]
# in order to support multiple content types...
class ChatMessageDict(TypedDict):
    content: str
    role: Role


class ChatMessage:
    def __init__(
        self,
        content: TagChild,
        role: Role = "assistant",
    ):
        self.role: Role = role

        is_html = isinstance(content, (Tag, TagList, HTML, Tagifiable))

        # content _can_ be a TagChild, but it's most likely just a string (of
        # markdown), so only process it if it's not a string.
        deps = []
        if not isinstance(content, str):
            session = get_current_session()
            if session and not session.is_stub_session():
                res = session._process_ui(content)
                deps = res["deps"]
            else:
                res = TagList(content).render()
                deps = [d.as_dict() for d in res["dependencies"]]
            content = res["html"]

        if is_html:
            # Code blocks with `{=html}` infostrings are rendered as-is by a
            # custom rendering method in markdown-stream.ts
            content = f"\n\n````````{{=html}}\n{content}\n````````\n\n"

        self.content = content
        self.html_deps = deps


# A message once transformed have been applied
@dataclass
class TransformedMessage:
    content_client: str | HTML
    content_server: str
    role: Role
    transform_key: Literal["content_client", "content_server"]
    pre_transform_key: Literal["content_client", "content_server"]
    html_deps: list[dict[str, str]] | None = None

    @classmethod
    def from_chat_message(cls, message: ChatMessage) -> "TransformedMessage":
        if message.role == "user":
            transform_key = "content_server"
            pre_transform_key = "content_client"
        else:
            transform_key = "content_client"
            pre_transform_key = "content_server"

        return TransformedMessage(
            content_client=message.content,
            content_server=message.content,
            role=message.role,
            transform_key=transform_key,
            pre_transform_key=pre_transform_key,
            html_deps=message.html_deps,
        )


# A message that can be sent to the client
class ClientMessage(TypedDict):
    content: str
    role: Literal["assistant", "user"]
    content_type: Literal["markdown", "html"]
    chunk_type: Literal["message_start", "message_end"] | None
    operation: Literal["append", "replace"]
    icon: NotRequired[str]
    html_deps: NotRequired[list[dict[str, str]]]



from __future__ import annotations

from typing import Literal, Pattern, Union

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

PatternStr = Pattern[str]
PatternOrStr = Union[str, PatternStr]
Timeout = Union[float, None]


# Avoid circular import with shiny.playwright by copy/pasting UiBase class
class UiBase:
    """A base class representing shiny UI components."""

    id: str
    """
    The browser DOM `id` of the UI element.
    """
    loc: Locator
    """
    Playwright `Locator` of the UI element.
    """
    page: Page
    """
    Playwright `Page` of the Shiny app.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: Locator | str,
    ) -> None:
        self.page = page
        # Needed?!? This is covered by `self.loc_root` and possibly `self.loc`
        self.id = id
        if isinstance(loc, str):
            loc = page.locator(loc)
        self.loc = loc

    @property
    def expect(self):
        """Expectation method equivalent to `playwright.expect(self.loc)`."""
        return playwright_expect(self.loc)


class Chat(UiBase):
    """Controller for :func:`shiny.ui.chat`."""

    loc: Locator
    """
    Playwright `Locator` for the chat.
    """
    loc_messages: Locator
    """
    Playwright `Locator` for the chat messages.
    """
    loc_latest_message: Locator
    """
    Playwright `Locator` for the last message in the chat.
    """
    loc_input_container: Locator
    """
    Playwright `Locator` for the chat input container.
    """
    loc_input: Locator
    """
    Playwright `Locator` for the chat's <textarea> input.
    """
    loc_input_button: Locator
    """
    Playwright `Locator` for the chat's <button> input.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Chat` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the chat.
        """
        super().__init__(
            page,
            id=id,
            loc=f"#{id}",
        )
        self.loc_messages = self.loc.locator("> shiny-chat-messages")
        self.loc_latest_message = self.loc_messages.locator("> :last-child")
        self.loc_input_container = self.loc.locator("> shiny-chat-input")
        self.loc_input = self.loc_input_container.locator("textarea")
        self.loc_input_button = self.loc_input_container.locator("button")

    def expect_latest_message(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the last message in the chat.

        Parameters
        ----------
        value
            The expected last message.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        # playwright_expect(self.loc_latest_message).to_have_text(value, timeout=timeout)
        playwright_expect(self.loc_latest_message).to_have_text(
            value, use_inner_text=True, timeout=timeout
        )

    def expect_messages(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the chat messages.

        Parameters
        ----------
        value
            The expected messages.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_messages).to_have_text(
            value, use_inner_text=True, timeout=timeout
        )

    def set_user_input(
        self,
        value: str,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Sets the user message in the chat.

        Parameters
        ----------
        value
            The message to send.
        timeout
            The maximum time to wait for the chat input to be visible and interactable. Defaults to `None`.
        """
        self.loc_input.type(value, timeout=timeout)

    def send_user_input(
        self,
        *,
        method: Literal["enter", "click"] = "enter",
        timeout: Timeout = None,
    ) -> None:
        """
        Sends the user message in the chat.

        Parameters
        ----------
        method
            The method to send the user message. Defaults to `"enter"`.
        timeout
            The maximum time to wait for the chat input to be visible and interactable. Defaults to `None`.
        """
        if method == "enter":
            self.loc_input.press("Enter", timeout=timeout)
        else:
            self.loc_input_button.click(timeout=timeout)

    def expect_user_input(
        self, value: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Expects the user message in the chat.

        Parameters
        ----------
        value
            The expected user message.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_input).to_have_value(value, timeout=timeout)



from __future__ import annotations

import asyncio
import contextlib
import functools
import inspect
import random
import secrets
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Generator,
    Iterable,
    TypeVar,
    cast,
)

from ._typing_extensions import ParamSpec, TypeGuard

CancelledError = asyncio.CancelledError


# ==============================================================================
# Misc utility functions
# ==============================================================================
def rand_hex(bytes: int) -> str:
    """
    Creates a random hexadecimal string of size `bytes`. The length in
    characters will be bytes*2.
    """
    format_str = "{{:0{}x}}".format(bytes * 2)
    return format_str.format(secrets.randbits(bytes * 8))


def drop_none(x: dict[str, Any]) -> dict[str, object]:
    return {k: v for k, v in x.items() if v is not None}


# Intended for use with json.load()'s object_hook parameter.
# Note also that object_hook is only called on dicts, not on lists, so this
# won't work for converting "top-level" lists to tuples
def lists_to_tuples(x: object) -> object:
    if isinstance(x, dict):
        x = cast("dict[str, object]", x)
        return {k: lists_to_tuples(v) for k, v in x.items()}
    elif isinstance(x, list):
        x = cast("list[object]", x)
        return tuple(lists_to_tuples(y) for y in x)
    else:
        # TODO: are there other mutable iterators that we want to make read only?
        return x


# ==============================================================================
# Private random stream
# ==============================================================================
def private_random_id(prefix: str = "", bytes: int = 3) -> str:
    if prefix != "" and not prefix.endswith("_"):
        prefix += "_"

    with private_seed():
        return prefix + rand_hex(bytes)


@contextlib.contextmanager
def private_seed() -> Generator[None, None, None]:
    state = random.getstate()
    global own_random_state  # noqa: PLW0603
    try:
        random.setstate(own_random_state)
        yield
    finally:
        own_random_state = random.getstate()
        random.setstate(state)


# Initialize random state for shiny's own private stream of randomness.
current_random_state = random.getstate()
random.seed(secrets.randbits(128))
own_random_state = random.getstate()
random.setstate(current_random_state)

# ==============================================================================
# Async-related functions
# ==============================================================================

R = TypeVar("R")  # Return type
P = ParamSpec("P")


def wrap_async(
    fn: Callable[P, R] | Callable[P, Awaitable[R]],
) -> Callable[P, Awaitable[R]]:
    """
    Given a synchronous function that returns R, return an async function that wraps the
    original function. If the input function is already async, then return it unchanged.
    """

    if is_async_callable(fn):
        return fn

    fn = cast(Callable[P, R], fn)

    @functools.wraps(fn)
    async def fn_async(*args: P.args, **kwargs: P.kwargs) -> R:
        return fn(*args, **kwargs)

    return fn_async


# This function should generally be used in this code base instead of
# `iscoroutinefunction()`.
def is_async_callable(
    obj: Callable[P, R] | Callable[P, Awaitable[R]],
) -> TypeGuard[Callable[P, Awaitable[R]]]:
    """
    Determine if an object is an async function.

    This is a more general version of `inspect.iscoroutinefunction()`, which only works
    on functions. This function works on any object that has a `__call__` method, such
    as a class instance.

    Returns
    -------
    :
        Returns True if `obj` is an `async def` function, or if it's an object with a
        `__call__` method which is an `async def` function.
    """
    if inspect.iscoroutinefunction(obj):
        return True
    if hasattr(obj, "__call__"):  # noqa: B004
        if inspect.iscoroutinefunction(obj.__call__):  # type: ignore
            return True

    return False


def wrap_async_iterable(
    x: Iterable[Any] | AsyncIterable[Any],
) -> AsyncIterable[Any]:
    """
    Given any iterable, return an async iterable. The async iterable will yield the
    values of the original iterable, but will also yield control to the event loop
    after each value. This is useful when you want to interleave processing with other
    tasks, or when you want to simulate an async iterable from a regular iterable.
    """

    if isinstance(x, AsyncIterable):
        return x

    if not isinstance(x, Iterable):
        raise TypeError("wrap_async_iterable requires an Iterable object.")

    return MakeIterableAsync(x)


class MakeIterableAsync:
    def __init__(self, iterable: Iterable[Any]):
        self.iterable = iterable

    def __aiter__(self):
        self.iterator = iter(self.iterable)
        return self

    async def __anext__(self):
        try:
            value = next(self.iterator)
            await asyncio.sleep(0)  # Yield control to the event loop
            return value
        except StopIteration:
            raise StopAsyncIteration



try:
    import playwright  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]
except ImportError:
    raise ImportError(
        "The shinychat.playwright module requires the playwright package to be installed. "
        "Please install it with: `pip install playwright`"
    )

# If `pytest` is installed...
try:
    import pytest  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]

    # At this point, `playwright` and `pytest` are installed.
    # Try to make sure `pytest-playwright` is installed
    try:
        import pytest_playwright  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]

    except ImportError:
        raise ImportError(
            "If you are using pytest to test your shiny app, install the pytest-playwright "
            "shim package with: `pip install pytest-playwright`"
        )
except ImportError:
    pass

from ._chat import Chat as ChatController

__all__ = ["ChatController"]
